
public class E.GSignonSessionOAuth : E.AuthenticationSession {
    Ag.Manager manager;

    construct {
        manager = new Ag.Manager ();
    }

    public override E.AuthenticationSessionResult execute_sync (GLib.Cancellable? cancellable = null) throws GLib.Error {
        var loop = new MainLoop ();
        E.AuthenticationSessionResult? result = null;
        Error error = null;
        execute.begin (GLib.Priority.DEFAULT, cancellable, (obj, res) => {
            try {
                result = (E.AuthenticationSessionResult)execute.end (res);
            } catch (Error e) {
                error = e;
            }

            loop.quit ();
        });

        loop.run ();
        if (error != null)
            throw error;

        if (result != null)
            return result;
        else
            return E.AuthenticationSessionResult.ERROR;
    }

    public override async E.AuthenticationSessionResult execute (int io_priority, GLib.Cancellable? cancellable = null) throws GLib.Error {
        var source = get_server ().ref_source (get_source_uid ());
        if (source == null) {
            warning ("No such data source for UID '%s'", get_source_uid ());
            throw new IOError.NOT_FOUND ("No such data source for UID '%s'", get_source_uid ());
        }

        try {
            var account_service = new_account_service (source);
            var auth_data = account_service.get_auth_data ();

            /* Hard-code the method and mechanism names.  If they disagree
             * with AgAuthData then hopefully the signon process will fail
             * with a suitable error message. */
            var identity = new Signon.Identity.from_db (auth_data.get_credentials_id ());
            var auth_session = identity.create_session ("oauth");
            auth_session.state_changed.connect ((session, state, msg) => {message (msg);});

            var mechanism = auth_data.get_mechanism ();
            string method;
            if (mechanism == "PLAINTEXT" || mechanism == "HMAC-SHA1" || mechanism == "RSA-SHA1") {
                method = "oauth1";
            } else if (mechanism == "web_server" || mechanism == "user_agent") {
                method = "oauth2";
            } else {
                method = mechanism;
            }
            var session_data = yield auth_session.process_async (auth_data.get_login_parameters (null), method, cancellable);
            var secret = session_data.lookup_value ("AccessToken", VariantType.STRING);
            if (secret == null) {
                throw new Signon.Error.MISSING_DATA ("Signon service did not return a secret");
            }

            var auth_result = yield get_authenticator ().try_password (new StringBuilder (secret.get_string ()), cancellable);
            if (auth_result == E.SourceAuthenticationResult.ACCEPTED) {
                return E.AuthenticationSessionResult.SUCCESS;
            }
        } catch (Error e) {
            throw e;
        }

        return E.AuthenticationSessionResult.ERROR;
    }

    private Ag.AccountService? new_account_service (E.Source source) throws GLib.Error {
        /* XXX The ESource should be a collection source with an
         *     [Ubuntu Online Accounts] extension.  Verify this. */
        return_val_if_fail (source.has_extension (E.SOURCE_EXTENSION_UOA), null);

        var extension = (E.SourceUoa)source.get_extension (E.SOURCE_EXTENSION_UOA);

        try {
            var account = manager.load_account (extension.get_account_id ());
            if (account == null)
                return null;

            /* XXX We can't accurately determine the appropriate service
             *     type from a collection source, but all services for an
             *     account should be using the same authentication method
             *     and mechanism so any service should work. */
            return new Ag.AccountService (account, account.list_services ().data);
        } catch (Error e) {
            throw e;
        }
    }
}

[ModuleInit]
public static Type e_fake_module_load (GLib.TypeModule type_module) {
    return typeof(E.GSignonSessionOAuth);
}
